#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/resource.h>
#include <math.h>
#include "interpd.h"
#include "utils.h"
#include "functions.h"

int main() {
	interplan plan;
	double *target;
	double xMin, xMax, yMin, yMax;
	int i, deg, tg_side, ntg;
	int interpRes;
	struct rusage tic, toc;
	struct timeval elapsed;
	double (*testfunct)(double, double);
	
	double fval, val, minVal, maxVal, meanVal, maxDev;
	double minErr, maxErr;
	
	deg = 150;
	tg_side = 100;
	ntg = tg_side * tg_side;
	
	xMin = 0.0;
	xMax = 1.0;
	yMin = 0.0;
	yMax = 1.0;
	
	testfunct = franke10;	
	
	generate_square_grid(tg_side, &target, xMin, xMax, yMin, yMax); 
	
	load_fftw_wisdom();
	
	getrusage(RUSAGE_SELF, &tic);
	init_plan(&plan, 1, deg, testfunct, xMin, xMax, yMin, yMax, 0);
	getrusage(RUSAGE_SELF, &toc);

	timeval_subtract(&elapsed, &(toc.ru_utime), &(tic.ru_utime));
	
	printf("Coefficients of degree %d created. Time: %ld.%06d s.\n", deg, elapsed.tv_sec, elapsed.tv_usec);
	
	getrusage(RUSAGE_SELF, &tic);
	interpRes = interpolate(&plan, ntg, target, 0, APPROX);
	getrusage(RUSAGE_SELF, &toc);
	
	timeval_subtract(&elapsed, &(toc.ru_utime), &(tic.ru_utime));
	
	printf("Function evaluated at %d target points. Time: %ld.%06d s.\n", ntg, elapsed.tv_sec, elapsed.tv_usec);
	
	maxErr = meanVal = 0;
	minErr = -HUGE_VAL;
	minVal = HUGE_VAL;
	maxVal = -HUGE_VAL;
	
	for (i = 0; i < ntg; i++) {
		fval = testfunct(target[2*i], target[2*i+1]);
		val = fabs(plan.funct_on_target[i] - fval);
		
		if (val > maxErr)
			maxErr = val;	
		if (val < minErr)
			minErr = val;
			
		maxVal = MAX(fval, maxVal);
		minVal = MIN(fval, minVal);
		meanVal += fval;		
	}
		
	meanVal = meanVal / (double)ntg;
	maxDev = MAX(meanVal - minVal, maxVal - meanVal);
	
	printf("Max normalized error: %.6e\n", maxErr/maxDev);

	
	plot_data(&plan, "data.dat");
	
	destroy_plan(&plan);
	
	return 0;
}
