#include <math.h>
#include <fftw3.h>
#include <nfft/nfft3.h>
#include <nfft/util.h>
#include BLASINC
#include "interpd.h"
#include "utils.h"

int func_on_targets(interplan *plan, unsigned flags) {
	/* plan->target must be Mx2 matrix */
	nfct_plan n_plan;
	int i, l, size;
	int nanErr = 0;
	
	size = plan->nr_coeff * plan->nc_coeff;
	
	if (plan->m > 0) {
		int dim[2];
		int oversmp[2];
	
		dim[0] = plan->nr_coeff;
		dim[1] = plan->nc_coeff;
		oversmp[0] = 2*(nfft_next_power_of_2(dim[0])-1);
		oversmp[1] = 2*(nfft_next_power_of_2(dim[1])-1);		
		
		nfct_init_guru(&n_plan, 2, dim, plan->ntarget, oversmp, plan->m,
			PRE_PHI_HUT | PRE_PSI | MALLOC_F_HAT | MALLOC_X | MALLOC_F | FFTW_INIT | FFT_OUT_OF_PLACE,
			FFTW_ESTIMATE | FFTW_DESTROY_INPUT);
	} else {
		nfct_init_2d(&n_plan, plan->nr_coeff, plan->nc_coeff, plan->ntarget);
	}


	/* We can save this directly saving the result of the fftw in the
	 * f_hat vector. */
	DCOPY(size, plan->coeff, 1, n_plan.f_hat, 1);
	
	// Set coefficients for normalized Chebyshev polynomials
	// All by 2 (we can save this by previously multiplying wstar by 2)
	DSCAL(size, 2.0, n_plan.f_hat, 1);
	
	// c(0,0) divided by 2
	n_plan.f_hat[0] = n_plan.f_hat[0] / 2.0;
	
	// first row divided by sqrt(2)
	DSCAL(plan->nc_coeff-1, INVSQRT2, n_plan.f_hat+1, 1);
	
	// first column divided by sqrt(2)
	DSCAL(plan->nr_coeff-1, INVSQRT2, n_plan.f_hat+plan->nc_coeff, plan->nc_coeff);
	
	
	l = 2*plan->ntarget;
	
	/* Multiply by 1/2*PI instead of dividing by 2*PI. This should use then
	 * the fused multiply-add instructions (where available). */
	for (i = 0; i < l; i++) {
		if (i % 2 != 0) {
			n_plan.x[i] = acos(DOM2PAD(plan->target[i], plan->yMin, plan->yMax)) * INVDOUBLEPI;
			if (isnan(n_plan.x[i]))
				nanErr = -1;
		}
		else {
			n_plan.x[i] = acos(DOM2PAD(plan->target[i], plan->xMin, plan->xMax)) * INVDOUBLEPI;
			if (isnan(n_plan.x[i]))
				nanErr = -1;
		}
	}
	
	if (flags == EXACT) {
		ndct_trafo(&n_plan);
	} else {
		if (n_plan.nfct_flags & PRE_ONE_PSI)
			nfct_precompute_psi(&n_plan);
		nfct_trafo(&n_plan);
	}

	plan->funct_on_target = (double *)fftw_malloc(sizeof(double)*plan->ntarget);
	DCOPY(plan->ntarget, n_plan.f, 1, plan->funct_on_target, 1);
	nfct_finalize(&n_plan);
	
	return nanErr;
}

int bl_func_on_targets(interplan *plan, int (*chebX)(int, double, double*), int (*chebY)(int, double, double*)) {
	int i, j, l;
	double *T1;
	double *T2;
	plan->funct_on_target = (double *)fftw_malloc(sizeof(double)*plan->ntarget);

	T1 = (double *)fftw_malloc(sizeof(double)*(plan->degree+1));
	T2 = (double *)fftw_malloc(sizeof(double)*(plan->degree+1));
	
	for (i = 0; i < plan->ntarget; i++) {
		chebX(plan->degree, DOM2PAD(plan->target[2*i],plan->xMin,plan->xMax), T1);
		chebY(plan->degree, DOM2PAD(plan->target[2*i+1],plan->yMin,plan->yMax), T2);
	
		l = 0;
		plan->funct_on_target[i] = 0.0;
	
		for (j = plan->degree; j >= 0; j--) {
			plan->funct_on_target[i] += T1[plan->degree-j]*DDOT(j+1,&(plan->coeff[l*plan->nc_coeff]),1,T2,1);
			l++;
		}
	}
	
	fftw_free(T1);
	fftw_free(T2);
	
	return 0;
}
