#include <stdlib.h>
#include <stdio.h>
#include <fftw3.h>
#include <time.h>
#include <sys/resource.h>
#include <math.h>
#ifdef NORMAL_BLAS
#include <cblas.h>
#elif ACML_BLAS
#include <acml.h>
#endif
#include "interpd.h"
#include "functions.h"
#include "utils.h"

extern void pdpts_(int *,char *,double *,double *,double *,int *);
extern void padua2_(int*,int*,int*,double*,double*,double*,double*,double*,double*,double*,double*);
extern double pd2val_(int*,int*,double*,double*,double*,double*,double*);

int main() {
	int i, j;
	int length;
	double *target, *pd2dtarget, *res;
	interplan plan;
	FILE *out, *err, *time;
	struct rusage tic;
	struct rusage toc;
	struct timeval elapsed;
	double maxErrInt, maxErrPad;
	double avgErrInt, avgErrPad;
	double minErrInt, minErrPad;
	double fval, maxVal, minVal, meanVal, maxDev;
	double val;
	double xMin, xMax, yMin, yMax;
	int interpRes;
	
	// Fortran declarations
	int deg = 30; // Set here the degree of interpolation.
	int degmax = 400;
	int npd;
	double *xi1;
	double *xi2;
	double *wxi;
	double *fpd;
	double *t1;
	double *t2;
	double *C0;
	double esterr;
	int size;
	double (*testfunct)(double, double);
	
	// Set here the test function.
	testfunct = franke1;
	// Set here the domain.
	xMin = 0.0;
	xMax = 1.0;
	yMin = 0.0;
	yMax = 1.0;
	
	int side[] = {33, 70, 100, 150, 200, 300, 350, 500, 600, 700, 850, 1000};
//	int side[] = {33, 70, 100, 150, 200, 300, 350, 500};
//	int degs[] = {30, 40, 50, 60, 70, 80, 90, 100, 150, 200, 250, 300, 350, 400};
	int degs[] = {30, 60, 90, 100, 120, 150, 180, 200};
	
	load_fftw_wisdom();

	size = (degmax+1)*(degmax+2)/2;
	xi1 = (double *)malloc(sizeof(double)*size);
	xi2 = (double *)malloc(sizeof(double)*size);
	wxi = (double *)malloc(sizeof(double)*size);
	fpd = (double *)malloc(sizeof(double)*size);
	t1 = (double *)malloc(sizeof(double)*size*(degmax+1));
	t2 = (double *)malloc(sizeof(double)*size*(degmax+1));
	C0 = (double *)malloc(sizeof(double)*size*2);
	
	out = fopen("tests", "w");
	err = fopen("test_error", "w");
	time = fopen("time", "w");
//	fprintf(out, "# Degree TimeP2D TimeIPD\n");
	fprintf(out, "Degree maxPad maxInt avgPad avgInt\n");
	fprintf(time, "Degree C0P2D C0IPD IntPD2 IntIPD\n");
	
	generate_square_grid(500, &target, xMin, xMax, yMin, yMax);
	generate_square_grid(500, &pd2dtarget, xMin, xMax, yMin, yMax);
	length = 250000;
	res = malloc(sizeof(double)*length);
	
	for (i = 0; i < length; i++) {
		pd2dtarget[2*i] = DOM2PAD(pd2dtarget[2*i], xMin, xMax);
		pd2dtarget[2*i+1] = DOM2PAD(pd2dtarget[2*i+1], yMin, yMax);
	}
	
	for (j = 0; j < 1; j++) {
		deg = degs[j];
		// Compute coefficients using Padua2D.	
		getrusage(RUSAGE_SELF, &tic);
		
		pdpts_(&deg, "TOP", xi1, xi2, wxi, &npd);
	
		for (i = 0; i < npd; i++)
			fpd[i] = testfunct(PAD2DOM(xi1[i],xMin,xMax), PAD2DOM(xi2[i],yMin,yMax));
	
		padua2_(&deg, &degmax, &npd, xi1, xi2, wxi, fpd, t1, t2, C0, &esterr);
		getrusage(RUSAGE_SELF, &toc);
		timeval_subtract(&elapsed, &(toc.ru_utime), &(tic.ru_utime));
		printf("Coefficients of degree %d created by Padua2D. Time: %ld.%06d s.\n", deg, elapsed.tv_sec, elapsed.tv_usec);
		fprintf(time, "%d %ld.%06d ", deg, elapsed.tv_sec, elapsed.tv_usec);
//		fprintf(out, "%d %ld.%06d ", deg, elapsed.tv_sec, elapsed.tv_usec);
	
		// Compute coefficients using InterPD.
		getrusage(RUSAGE_SELF, &tic);
		init_plan(&plan, 1, deg, testfunct, xMin, xMax, yMin, yMax, 0);
		getrusage(RUSAGE_SELF, &toc);
		timeval_subtract(&elapsed, &(toc.ru_utime), &(tic.ru_utime));
		printf("Coefficients of degree %d created by InterPD. Time: %ld.%06d s.\n", deg, elapsed.tv_sec, elapsed.tv_usec);
//		fprintf(out, "%ld.%06d\n", elapsed.tv_sec, elapsed.tv_usec);
		fprintf(time, "%ld.%06d ", elapsed.tv_sec, elapsed.tv_usec);

		// Compute interpolation using Padua2D.
		getrusage(RUSAGE_SELF, &tic);
	
		for (i = 0; i < length; i++) {
			res[i] = pd2val_(&deg, &degmax, C0, &pd2dtarget[2*i], &pd2dtarget[2*i+1], t1, t2);
		}
	
		getrusage(RUSAGE_SELF, &toc);
	
		timeval_subtract(&elapsed, &(toc.ru_utime), &(tic.ru_utime));

		printf("Function on target by Padua2D. Time: %ld.%06d s.\n", elapsed.tv_sec, elapsed.tv_usec);
//		fprintf(out, "%ld.%06d ", elapsed.tv_sec, elapsed.tv_usec);
		fprintf(time, "%ld.%06d ", elapsed.tv_sec, elapsed.tv_usec);

		// Compute interpolation using InterPD.
		getrusage(RUSAGE_SELF, &tic);
		interpRes = interpolate(&plan, length, target, APPROX);
		getrusage(RUSAGE_SELF, &toc);
		
		timeval_subtract(&elapsed, &(toc.ru_utime), &(tic.ru_utime));
		printf("Function on target by InterPD. Time: %ld.%06d s.\n", elapsed.tv_sec, elapsed.tv_usec);
		fprintf(time, "%ld.%06d\n", elapsed.tv_sec, elapsed.tv_usec);
		
		if (interpRes < 0) {
			printf("################## NAN DETECTED! ##################\n");
			fprintf(err, "NaN detected!\n");
		}
		
		// Compute the max error.
		maxErrInt = maxErrPad = avgErrInt = avgErrPad = meanVal = 0.0;
		minErrInt = minErrPad = HUGE_VAL; // from math.h
		minVal = HUGE_VAL;
		maxVal = -HUGE_VAL;
	
		for (i = 0; i < length; i++) {
			fval = testfunct(target[2*i], target[2*i+1]);
			val = fabs(plan.funct_on_target[i] - fval);
			if (val > maxErrInt) {
//				fprintf(err, "New max found %.6e: interpolated %.6e should be %.6e.\n", val, plan.funct_on_target[i], testfunct(target[2*i], target[2*i+1]));
//				fprintf(err, "---> Target values: x=%.12e y=%.12e. Target index: %d.\n", target[2*i], target[2*i+1], i);
//				fprintf(err, "----> Mapped to x=%.6e y=%.6e\n", DOM2PAD(target[2*i],xMin,xMax), DOM2PAD(target[2*i+1],yMin,yMax));
				maxErrInt = val;
			}
			if (val < minErrInt)
				minErrInt = val;
			
			maxVal = MAX(fval, maxVal);
			minVal = MIN(fval, minVal);
			meanVal += fval;
			
			avgErrInt += val;
		
			val = fabs(res[i] - fval);
			if (val > maxErrPad)
				maxErrPad = val;
			if (val < minErrPad)
				minErrPad = val;
			
			avgErrPad += val;
		}
		
//		fprintf(err, "================================================================================\n\n");
		avgErrInt = avgErrInt / (double)length;
		avgErrPad = avgErrPad / (double)length;
		meanVal = meanVal / (double)length;
		maxDev = MAX(meanVal - minVal, maxVal - meanVal);
		
		plot_data(&plan, NULL);
		destroy_plan(&plan);
		fprintf(out, "%d %.6e %.6e %.6e %.6e\n", deg, maxErrPad/maxDev, maxErrInt/maxDev, avgErrPad, avgErrInt);
		printf("Max normalized error by Padua2D: %.6e\n", maxErrPad/maxDev);
		printf("Max normalized error by InterPD: %.6e\n", maxErrInt/maxDev);
		printf("Average error by Padua2D: %.6e\n", avgErrPad);
		printf("Average error by InterPD: %.6e\n", avgErrInt);
		printf("Max deviation by Padua2D: %.6e\n", MAX(maxErrPad-avgErrPad, avgErrPad-minErrPad));
		printf("Max deviation by InterPD: %.6e\n\n", MAX(maxErrInt-avgErrInt, avgErrInt-minErrInt)); 
		
	}

//	out = fopen("tests", "w");
//	err = fopen("test_error", "w");
//	fprintf(out, "# Interpolation of degree %d at Padua points of 1 family\n", deg);
//	fprintf(out, "#N points TimeP2D TimeIPD ErrP2D ErrIPD AvgErrP2D AvgErrIPD\n");
/*	
	for (j = 0; j < 12; j++) {
		// Range per errore 10e-1: 161-297 (????)
		printf("%dx%d target points:\n", side[j], side[j]);
		generate_square_grid(side[j], &target, xMin, xMax, yMin, yMax);
		generate_square_grid(side[j], &pd2dtarget, xMin, xMax, yMin, yMax);
		length = side[j]*side[j];
		fprintf(out, "%d ", length);
		fprintf(err, "Degree: %d Number of target points: %d.\n", deg, length);
		res = (double *)malloc(sizeof(double)*length);
		//generate_random(length, &target, 0.0, 1.0, 0.0, 1.0);
		//pd2dtarget = (double *)fftw_malloc(sizeof(double)*2*length);
		//cblas_dcopy(2*length, target, 1, pd2dtarget, 1);
		// Compute coefficients using Padua2D.	
		getrusage(RUSAGE_SELF, &tic);
		
		pdpts_(&deg, "TOP", xi1, xi2, wxi, &npd);
	
		for (i = 0; i < npd; i++)
			fpd[i] = testfunct(PAD2DOM(xi1[i],xMin,xMax), PAD2DOM(xi2[i],yMin,yMax));
	
		padua2_(&deg, &degmax, &npd, xi1, xi2, wxi, fpd, t1, t2, C0, &esterr);
		getrusage(RUSAGE_SELF, &toc);
		timeval_subtract(&elapsed, &(toc.ru_utime), &(tic.ru_utime));
		printf("Coefficients created by Padua2D. Time: %ld.%06d s.\n", elapsed.tv_sec, elapsed.tv_usec);
	
		// Compute coefficients using InterPD.
		getrusage(RUSAGE_SELF, &tic);
		init_plan(&plan, 1, deg, testfunct, xMin, xMax, yMin, yMax, 0);
		getrusage(RUSAGE_SELF, &toc);
		timeval_subtract(&elapsed, &(toc.ru_utime), &(tic.ru_utime));
		printf("Coefficients created by InterPD. Time: %ld.%06d s.\n", elapsed.tv_sec, elapsed.tv_usec);
		
		for (i = 0; i < length; i++) {
			pd2dtarget[2*i] = DOM2PAD(pd2dtarget[2*i], plan.xMin, plan.xMax);
			pd2dtarget[2*i+1] = DOM2PAD(pd2dtarget[2*i+1], plan.yMin, plan.yMax);
		}

		// Compute interpolation using Padua2D.
		getrusage(RUSAGE_SELF, &tic);
	
		for (i = 0; i < length; i++) {
			res[i] = pd2val_(&deg, &degmax, C0, &pd2dtarget[2*i], &pd2dtarget[2*i+1], t1, t2);
		}
	
		getrusage(RUSAGE_SELF, &toc);
	
		timeval_subtract(&elapsed, &(toc.ru_utime), &(tic.ru_utime));

		printf("Function on target by Padua2D. Time: %ld.%06d s.\n", elapsed.tv_sec, elapsed.tv_usec);
		fprintf(out, "%ld.%06d ", elapsed.tv_sec, elapsed.tv_usec);

		// Compute interpolation using InterPD.
		getrusage(RUSAGE_SELF, &tic);
		interpRes = interpolate(&plan, length, target, APPROX);
		getrusage(RUSAGE_SELF, &toc);
		
		if (interpRes < 0) {
			printf("################## NAN DETECTED! ##################\n");
			fprintf(err, "NaN detected!\n");
		}
	
//		fprintf(out, "# Interpolated function at %d target points.\n", length);
	
//		for (i = 0; i < length; i++)
//			fprintf(out, "%.8E %.8E %.8E\n", target[2*i], target[2*i+1], plan.funct_on_target[i]);

		timeval_subtract(&elapsed, &(toc.ru_utime), &(tic.ru_utime));
		printf("Function on target by InterPD. Time: %ld.%06d s.\n", elapsed.tv_sec, elapsed.tv_usec);
		fprintf(out, "%ld.%06d ", elapsed.tv_sec, elapsed.tv_usec);
	
		// Compute the max error.
		maxErrInt = maxErrPad = avgErrInt = avgErrPad = 0.0;
		minErrInt = minErrPad = HUGE_VAL; // from math.h
	
		for (i = 0; i < length; i++) {
			val = fabs(plan.funct_on_target[i] - testfunct(target[2*i], target[2*i+1]));
			if (val > maxErrInt) {
				fprintf(err, "New max found %.6e: interpolated %.6e should be %.6e.\n", val, plan.funct_on_target[i], testfunct(target[2*i], target[2*i+1]));
				fprintf(err, "---> Target values: x=%.12e y=%.12e. Target index: %d.\n", target[2*i], target[2*i+1], i);
				fprintf(err, "----> Mapped to x=%.6e y=%.6e\n", DOM2PAD(target[2*i],xMin,xMax), DOM2PAD(target[2*i+1],yMin,yMax));
				maxErrInt = val;
			}
			if (val < minErrInt)
				minErrInt = val;
			
			avgErrInt += val;
		
			val = fabs(res[i] - testfunct(target[2*i], target[2*i+1]));
			if (val > maxErrPad)
				maxErrPad = val;
			if (val < minErrPad)
				minErrPad = val;
			
			avgErrPad += val;
		}
		
		fprintf(err, "================================================================================\n\n");
		avgErrInt = avgErrInt / (double)length;
		avgErrPad = avgErrPad / (double)length;
		free(res);
		
		destroy_plan(&plan);
		fftw_free(target);
		fftw_free(pd2dtarget);
		fprintf(out, "%.6e %.6e %.6e %.6e\n", maxErrPad, maxErrInt, avgErrPad, avgErrInt);
		printf("Max error by Padua2D: %.6e\n", maxErrPad);
		printf("Max error by InterPD: %.6e\n", maxErrInt);
		printf("Average error by Padua2D: %.6e\n", avgErrPad);
		printf("Average error by InterPD: %.6e\n", avgErrInt);
		printf("Max deviation by Padua2D: %.6e\n", MAX(maxErrPad-avgErrPad, avgErrPad-minErrPad));
		printf("Max deviation by InterPD: %.6e\n\n", MAX(maxErrInt-avgErrInt, avgErrInt-minErrInt)); 

	}
*/	
	fclose(out);
//	fclose(err);
		
	return 0;
}
