#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <fftw3.h>
#include "interpd.h"
#include "scatter.h"

// Functions of CSHEP
extern void cshep2_(int *N,double *X,double *Y,double *F,int *NC,int *NW,int *NR,int *LCELL,int *LNEXT,double *XMIN,
					double *YMIN,double *DX,double *DY,double *RMAX,double *RW,double *A,int *IER);

extern double cs2val_(double *PX,double *PY,int *N,double *X,double *Y,double *F,int *NR,
					int *LCELL,int *LNEXT,double *XMIN,double *YMIN,double *DX,double *DY,double *RMAX,double *RW,double *A);


int init_scatter(scatter_pl *sc, double *scx, double *scy, double *v, int nv) {
	int ier;
	
	if (scx == NULL || scy == NULL || v == NULL || sc == NULL)
		return -1;
	if (nv < 1)
		return -1;
	
	sc->nc = 17;
	sc->nw = 30;
	sc->nr = lround(sqrt(nv/3.0));
	sc->lcell = (int *)malloc(sizeof(int)*sc->nr*sc->nr);
	sc->lnext = (int *)malloc(sizeof(int)*nv);
	sc->A = (double *)malloc(sizeof(double)*9*nv);
	sc->rw = (double *)malloc(sizeof(double)*nv);
	
	sc->X = scx;
	sc->Y = scy;
	sc->F = v;
	sc->N = nv;
	
	cshep2_(&(sc->N),sc->X,sc->Y,sc->F,&(sc->nc),&(sc->nw),&(sc->nr),sc->lcell,sc->lnext,&(sc->xmin),&(sc->ymin),&(sc->dx),&(sc->dy),&(sc->rmax),sc->rw,sc->A,&ier);
	
	return ier;
}

int interp_to_pd(scatter_pl *sc, interplan *plan, int f, int d, double xMin, double xMax, double yMin, double yMax, char *filename) {
	double *xis1, *xis2;
	int i, j, size;
	double wstar;
	int ok_save = 0;

	FILE *out;
	double fval;
	
	if (plan == NULL || sc == NULL)
		return -1;
	if (f < 1 || f > 4 || d < 1)
		return -1;
	
	if (filename != NULL) {
		out = fopen(filename, "w");
		if (out != NULL)
			ok_save = 1;
	}

	
	plan->family = f;
	plan->degree = d;
	plan->xMin = xMin;
	plan->xMax = xMax;
	plan->yMin = yMin;
	plan->yMax = yMax;
	plan->fftw_flags = FFTW_ESTIMATE;
	
	if (ok_save == 1) {
		fprintf(out, "%d\n", f);
		fprintf(out, "%d\n", d);
		fprintf(out, "%.26e\n", xMin);
		fprintf(out, "%.26e\n", xMax);
		fprintf(out, "%.26e\n", yMin);
		fprintf(out, "%.26e\n", yMax);
	}
	
	if (gen_xi_star(plan, &xis1, &xis2) < 0)
		return -1;
	
	size = (d+1)*(d+2);
	plan->fihw = (double *)fftw_malloc(size*sizeof(double));
	
	/* Set to zero the whole matrix vector. Blas might be faster than a for loop,
	 * because it may use a machine dependent SIMD instruction set. */
	DSCAL(size, 0.0, plan->fihw, 1);
	
	switch (plan->family) {
		case 1:
		for (i = 0; i <= plan->degree; i++)
			for (j = 0; j <= plan->degree+1; j++)
				if (!same_parity(i, j)) {
					fval = cs2val_(&xis1[i],&xis2[j],&(sc->N),sc->X,sc->Y,sc->F,&(sc->nr),sc->lcell,sc->lnext,&(sc->xmin),&(sc->ymin),&(sc->dx),&(sc->dy),&(sc->rmax),sc->rw,sc->A);
					if (ok_save == 1)
						fprintf(out, "%.26e\n", fval);
					set_element(plan->fihw, i, j, plan->degree+2, fval);
				}
		break;
		
		case 2:
		for (i = 0; i <= plan->degree+1; i++)
			for (j = 0; j <= plan->degree; j++)
				if (!same_parity(i, j)) {
					fval = cs2val_(&xis1[i],&xis2[j],&(sc->N),sc->X,sc->Y,sc->F,&(sc->nr),sc->lcell,sc->lnext,&(sc->xmin),&(sc->ymin),&(sc->dx),&(sc->dy),&(sc->rmax),sc->rw,sc->A);
					if (ok_save == 1)
						fprintf(out, "%.26e\n", fval);
					set_element(plan->fihw, i, j, plan->degree+1, fval);
				}
		break;
		
		case 3:
		for (i = 0; i <= plan->degree; i++)
			for (j = 0; j <= plan->degree+1; j++)
				if (same_parity(i, j)) {
					fval = cs2val_(&xis1[i],&xis2[j],&(sc->N),sc->X,sc->Y,sc->F,&(sc->nr),sc->lcell,sc->lnext,&(sc->xmin),&(sc->ymin),&(sc->dx),&(sc->dy),&(sc->rmax),sc->rw,sc->A);
					if (ok_save == 1)
						fprintf(out, "%.26e\n", fval);
					set_element(plan->fihw, i, j, plan->degree+2, fval);
				}
		break;
		
		case 4:
		for (i = 0; i <= plan->degree+1; i++)
			for (j = 0; j <= plan->degree; j++)
				if (same_parity(i, j)) {
					fval = cs2val_(&xis1[i],&xis2[j],&(sc->N),sc->X,sc->Y,sc->F,&(sc->nr),sc->lcell,sc->lnext,&(sc->xmin),&(sc->ymin),&(sc->dx),&(sc->dy),&(sc->rmax),sc->rw,sc->A);
					if (ok_save == 1)
						fprintf(out, "%.26e\n", fval);
					set_element(plan->fihw, i, j, plan->degree+1, fval);
				}
		break;
		
		default: // can't be, unless the memory is broken
		break;
	}
	
	// free the memory.
	free(xis1);
	free(xis2);
	
	wstar = 1.0 / (d*(d+1));
	DSCAL(size, wstar, plan->fihw, 1);
	
	gen_coeff(plan);
	
	if (ok_save == 1)
		fclose(out);
	
	return 0;
}

int end_scatter(scatter_pl *sc) {
	free(sc->lcell);
	free(sc->lnext);
	free(sc->A);
	free(sc->rw);
	return 0;
}

int load_scatter(interplan *plan, char *filename) {
	char str[1024];
	char *endptr;
	FILE *in;
	int i, j, size;
	double fval, wstar;
	
	in = fopen(filename, "r");
	
	if (in == NULL)
		return -1;
	
	if (fgets(str, 1020, in) == NULL) {
		fclose(in);
		return -1;
	}
	plan->family = strtol(str, &endptr, 10);
	
	if (endptr == str || plan->family < 1 || plan->family > 4) {
		fclose(in);
		return -1;
	}
	
	if (fgets(str, 1020, in) == NULL) {
		fclose(in);
		return -1;
	}
	plan->degree = strtol(str, &endptr, 10);
	
	if (endptr == str || plan->degree < 1) {
		fclose(in);
		return -1;
	}
	
	if (fgets(str, 1020, in) == NULL) {
		fclose(in);
		return -1;
	}
	plan->xMin = strtod(str, &endptr);
	
	if (endptr == str) {
		fclose(in);
		return -1;
	}
	
	if (fgets(str, 1020, in) == NULL) {
		fclose(in);
		return -1;
	}
	plan->xMax = strtod(str, &endptr);
	
	if (endptr == str) {
		fclose(in);
		return -1;
	}
	
	if (fgets(str, 1020, in) == NULL) {
		fclose(in);
		return -1;
	}
	plan->yMin = strtod(str, &endptr);
	
	if (endptr == str) {
		fclose(in);
		return -1;
	}
	
	if (fgets(str, 1020, in) == NULL) {
		fclose(in);
		return -1;
	}
	plan->yMax = strtod(str, &endptr);
	
	if (endptr == str) {
		fclose(in);
		return -1;
	}
	
	size = (plan->degree+1) * (plan->degree+2);
	plan->fihw = (double *)fftw_malloc(sizeof(double)*size);
	DSCAL(size, 0.0, plan->fihw, 1);
	
	switch (plan->family) {
		case 1:
		for (i = 0; i <= plan->degree; i++)
			for (j = 0; j <= plan->degree+1; j++)
				if (!same_parity(i, j)) {
					if (fgets(str, 1020, in) == NULL) {
						fclose(in);
						fftw_free(plan->fihw);
						perror("Stored data do not agree with degree and family.");
						return -1;
					}
					fval = strtod(str, &endptr);
					if (str == endptr) {
						fclose(in);
						fftw_free(plan->fihw);
						return -1;
					}
					set_element(plan->fihw, i, j, plan->degree+2, fval);
				}
		break;
		
		case 2:
		for (i = 0; i <= plan->degree+1; i++)
			for (j = 0; j <= plan->degree; j++)
				if (!same_parity(i, j)) {
					if (fgets(str, 1020, in) == NULL) {
						fclose(in);
						fftw_free(plan->fihw);
						perror("Stored data do not agree with degree and family.");
						return -1;
					}
					fval = strtod(str, &endptr);
					if (str == endptr) {
						fclose(in);
						fftw_free(plan->fihw);
						return -1;
					}
					set_element(plan->fihw, i, j, plan->degree+1, fval);
				}
		break;
		
		case 3:
		for (i = 0; i <= plan->degree; i++)
			for (j = 0; j <= plan->degree+1; j++)
				if (same_parity(i, j)) {
					if (fgets(str, 1020, in) == NULL) {
						fclose(in);
						fftw_free(plan->fihw);
						perror("Stored data do not agree with degree and family.");
						return -1;
					}
					fval = strtod(str, &endptr);
					if (str == endptr) {
						fclose(in);
						fftw_free(plan->fihw);
						return -1;
					}
					set_element(plan->fihw, i, j, plan->degree+2, fval);
				}
		break;
		
		case 4:
		for (i = 0; i <= plan->degree+1; i++)
			for (j = 0; j <= plan->degree; j++)
				if (same_parity(i, j)) {
					if (fgets(str, 1020, in) == NULL) {
						fclose(in);
						fftw_free(plan->fihw);
						perror("Stored data do not agree with degree and family.");
						return -1;
					}
					fval = strtod(str, &endptr);
					if (str == endptr) {
						fclose(in);
						fftw_free(plan->fihw);
						return -1;
					}
					set_element(plan->fihw, i, j, plan->degree+1, fval);
				}
		break;
		
		default: // can't be, unless the memory is broken
		break;
	}
	
	fclose(in);
	
	wstar = 1.0 / (plan->degree*(plan->degree+1));
	DSCAL(size, wstar, plan->fihw, 1);
	
	gen_coeff(plan);
	
	return 0;
}
