#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <nfft/nfft3.h>
#include "utils.h"
#include "interpd.h"
#include "functions.h"
#include "scatter.h"

int main() {
	double *scatter;
	double *scx, *scy, *scv;
	interplan plan, plan2;
	scatter_pl sc;
	int nsc = 5000;
	int deg[] = {16, 24, 32, 40, 48};
	int i,j;
	double t;
	
	double maxVal, minVal, meanVal, val;
	double maxErr, maxDev, error;
	
//	FILE *tst;
	
//	tst = fopen("errtst", "w");
//	fprintf(tst, "# Error at %d random points.\n", nsc);
	
	generate_random(nsc, &scatter, 0,1, 0,1);
	
	scx = (double *)malloc(sizeof(double)*nsc);
	scy = (double *)malloc(sizeof(double)*nsc);
	scv = (double *)malloc(sizeof(double)*nsc);
	
	for (i = 0; i < nsc; i++) {
		scx[i] = scatter[2*i];
		scy[i] = scatter[2*i+1];
		scv[i] = franke1(scx[i], scy[i]);
	}
	
	t = nfft_second();
	init_scatter(&sc, scx, scy, scv, nsc);
	t = nfft_second() - t;
	printf("Building CSHEP2 interpolant: %.6f s.\n", t);
	
	for (j = 0; j < 5; j++) {
		error = HUGE_VAL;
		
		t = nfft_second();
		if (interp_to_pd(&sc, &plan, 1, deg[j], 0,1, 0,1,NULL) != 0) {
			printf("Error!\n");
			return -1;
		}
		t = nfft_second() - t;
		printf("CSHEP2 time: %.6f s.\n", t);
		interpolate_bl(&plan, nsc, scatter);
	
		maxErr = 0.0;
		maxVal = -HUGE_VAL;
		minVal = HUGE_VAL;
		meanVal = 0.0;
	
		for (i = 0; i < nsc; i++) {
			val = fabs(plan.funct_on_target[i] - scv[i]);
		
			if (val > maxErr)
				maxErr = val;
			if (scv[i] > maxVal)
				maxVal = scv[i];
			if (scv[i] < minVal)
				minVal = scv[i];
		
			meanVal += scv[i];
		}
	
		meanVal = meanVal / (double)nsc;
		maxDev = MAX(meanVal - minVal, maxVal - meanVal);
	
		error = maxErr/maxDev;
		printf("Error with degree %d: %.3E\n", deg[j], error);
//		fprintf(tst, "%d %.2E\n", deg[j], error);
		
//		if (deg[j] == 40)
//			save_data(&plan, "scatter.dat");
	
		destroy_plan(&plan);
	}
/*
	plot_data(&plan, NULL);
	
	interp_to_pd(&sc, &plan, 1, deg, 0,1, 0,1,"padua.dat");
	
	load_scatter(&plan2, "padua.dat");
	interpolate_bl(&plan2, nsc, scatter);
	plot_data(&plan2, NULL);
*/	
//	fclose(tst);
	return 0;
}
